(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml + CDuce                    *)
(*                                                                     *)
(*            Alain Frisch, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typeext.ml,v 1.30 2005/04/18 13:50:30 frisch Exp $ *)

(* Typechecking for the CDuce extension *)

open Parsetree
open Types
open Btype
open Ctype

module CT = Cduce_types.Types
module SEQ = Cduce_types.Sequence
module U = Cduce_types.Encodings.Utf8
module Ns = Cduce_types.Ns
module CEnv = Cduce_types.Ident.Env
module IdSet = Cduce_types.Ident.IdSet
module Patterns = Cduce_types.Patterns
module LabelMap = Cduce_types.Ident.LabelMap
module LabelPool = Cduce_types.Ident.LabelPool

let real_type_expect = ref (fun ?in_function _ _ _ -> assert false)
let transl_simple_type = ref None

let solve_env = ref Env.empty
let real_repr t = repr (expand_head !solve_env t)

type t = CT.t

let id x = Cduce_types.Ident.ident (Cduce_types.Ns.empty, U.mk_latin1 x)

type error = 
  | UnboundNamespacePrefix of string
  | DuplicatedLabel
  | CaptureNotAllowed of string
  | UnguardedRecursion
  | NotXmlType
  | EmptyType
  | Cyclic
  | NotSubtype of t * t
  | CannotTranslateML of Types.type_expr
  | AddMix of t

exception Error of Location.t * error

let register_exttypes () =
  let types = Env.exttypes real_repr !solve_env in
  List.iter 
    (fun (p,ty) ->
       CT.Print.register_global
	 CT.CompUnit.pervasives
	 (Cduce_types.Ns.empty,
	  Cduce_types.Encodings.Utf8.mk_latin1 p)
	 ty) types
  
let report_error ppf = function
  | UnboundNamespacePrefix pr ->
      Format.fprintf ppf "Unbound namespace prefix %s" pr
  | DuplicatedLabel ->
      Format.fprintf ppf "The same label appears several times in this record"
  | CaptureNotAllowed s ->
      Format.fprintf ppf "Capture variable not allowed in types: %s" s
  | UnguardedRecursion ->
      Format.fprintf ppf "Unguarded recursion"
  | NotXmlType ->
      Format.fprintf ppf "Not an XML type"
  | EmptyType ->
      Format.fprintf ppf "This definition yields an empty type"
  | NotSubtype (t1,t2) ->
      register_exttypes ();
      Format.fprintf ppf "Subtyping failed %a <= %a@."
	CT.Print.print t1
	CT.Print.print t2;
      Format.fprintf ppf "Sample:@.%a"
        Cduce_types.Sample.print
	(Cduce_types.Sample.get (CT.diff t1 t2))
  | Cyclic ->
      Format.fprintf ppf "Cycle detected: cannot type-check"
  | CannotTranslateML t ->
      Format.fprintf ppf "Cannot translate ML type: %a"
	Printtyp.type_expr t
  | AddMix t ->
      register_exttypes ();
      Format.fprintf ppf "Left argument of + is not subtype of Int or Record: %a@."
	CT.Print.print t

let error loc err = raise (Error (loc,err))

let noloc = Location.none

(* Built-in types *)

let builtin_types = [ 
  "Int" , CT.Int.any;
  "Char" , CT.Char.any;
  "Any" , CT.any;
  "Empty" , CT.empty;
  "Atom" , CT.Atom.any;
  "String" , SEQ.string;
  "Latin1" , SEQ.string_latin1;
  "Int32" , (CT.interval Cduce_types.Intervals.int32);
  "Int64" , (CT.interval Cduce_types.Intervals.int64);
  "Arrow",  CT.Arrow.any;
  "Latin1Char", SEQ.char_latin1;
]

let built_ins =
  List.fold_left (
    fun accu (name,t) ->
      CT.Print.register_global CT.CompUnit.pervasives 
	(Ns.empty, U.mk_latin1 name) t;
      CEnv.add (id name) t accu
  ) CEnv.empty builtin_types

(* Solver *)


let anyext = { ext_const = None; ext_lb = []; ext_atoms = [] }

let check loc t ub =
  if CT.subtype t ub then t
  else error loc (NotSubtype (t,ub))

let compute_atom loc ub = function
  | { ext_atom_v = Atom_computed t } -> 
      check loc t ub
  | { ext_atom_v = Atom_computing } -> 
      error loc Cyclic
  | { ext_atom_v = Atom_start } as a -> 
      a.ext_atom_v <- Atom_computing; 
      let t = a.ext_atom_def ub in a.ext_atom_v <- Atom_computed t; 
      check loc t ub

let var_atoms v e ub =
  let atoms = ref [] and vs = ref [v] and consts = ref CT.empty in
  let rec aux e =  atoms := e.ext_atoms @ !atoms; List.iter aux2 e.ext_lb
  and aux2 (loc,v) =
    let v = real_repr v in
    if (List.memq v !vs) then () 
    else (vs := v :: !vs; match v.desc with
	    | Text { ext_const = Some t } -> 
		consts := CT.cup !consts (check loc t ub)
	    | Text e -> aux e
	    | _ -> assert false)
  in
  aux e;
  !atoms,!consts
  
let real_compute_var v e ub =
  let atoms,csts = var_atoms v e ub in
  List.fold_left 
    (fun accu (loc,at) -> CT.cup accu (compute_atom loc ub at)) 
    csts atoms

let compute_var loc v ub =
  let v = real_repr v in
  match v.desc with
    | Text { ext_const = Some t } -> check loc t ub
    | Text ({ ext_const = None } as e) -> 
	let t = real_compute_var v e ub in 
	v.desc <- Text { anyext with ext_const = Some t };
	t
    | _ -> assert false

let check_const v =
  match v.desc with
    | Text ({ ext_const = Some t } as e) ->
	ignore (real_compute_var v e t);
	v.desc <- Text { anyext with ext_const = Some t }
    | _ -> ()


let solve env =
  solve_env := env;
  let vars = List.map real_repr (allextvars ()) in
  List.iter check_const vars;
  List.iter (fun v -> ignore (compute_var Location.none v CT.any)) vars;
  register_exttypes ()


(* From ML types to CDuce types *)

let rec seq = function
  | [] -> SEQ.nil_node
  | hd::tl -> CT.cons (CT.times hd (seq tl))
  
let ml_constr c p =
  let c = Cduce_types.Atoms.V.mk Ns.empty (U.mk_latin1 c) in
  let c = CT.atom (Cduce_types.Atoms.atom c) in
  match p with
    | [] -> c
    | p ->
	let p = seq (CT.cons (CT.record_fields (false,LabelMap.empty)) :: p) in
	CT.xml (CT.cons c) p

type from_ml_env = { loc : Location.t; 
		     seen : type_expr list;
		     constrs : (Path.t * (type_expr list) * CT.Node.t) list }

let rec typ_from_ml env t =
  let t = real_repr t in
  if List.memq t env.seen then raise Exit
  else let env = { env with seen = t :: env.seen } in
  match t.desc with
    | Tconstr (p,_,_) when Path.same p Predef.path_unit ->
	SEQ.nil_node
    | Tconstr (p,_,_) when Path.same p Predef.path_int ->
	CT.cons (CT.Int.any)
    | Tconstr (p,_,_) when Path.same p Predef.path_int32 ->
	CT.cons (CT.interval Cduce_types.Intervals.int32)
    | Tconstr (p,_,_) when Path.same p Predef.path_int64 ->
	CT.cons (CT.interval Cduce_types.Intervals.int64)
    | Tconstr (p,_,_) when Path.same p Predef.path_char ->
	CT.cons SEQ.char_latin1
    | Tconstr (p,_,_) when Path.same p Predef.path_string ->
	CT.cons SEQ.string_latin1
    | Tconstr (p,_,_) when Path.same p Predef.path_utf8 ->
	CT.cons SEQ.string
    | Tconstr (p,[t],_) when Path.same p Predef.path_list ->
	SEQ.star_node (typ_from_ml env t)
    | Tconstr (p,_,_) when Path.name p = "Cduce_types.Value.t" ->
	CT.any_node
    | Tconstr (p,_,_) when Path.same p Predef.path_exn ->
	raise Exit
    | Ttuple tl ->
	CT.cons (CT.tuple (List.map (typ_from_ml env) tl))
    | Text _ ->
	CT.cons (compute_var env.loc t CT.any)
    | Tconstr (p,args,_) ->
	typ_from_ml_constr env p args
    | Tarrow (_,t1,t2,_) ->
	CT.cons (CT.arrow (typ_from_ml env t1) (typ_from_ml env t2))
    | Tvariant rd ->
	let rd = Btype.row_repr rd in
	if not rd.row_closed then raise Exit;
	let fields = Ctype.get_fields rd in
	CT.cons 
	  (List.fold_left
	     (fun accu (lab,f) -> 
		CT.cup accu (ml_constr lab (match f with None -> []
					      | Some t -> [typ_from_ml env t]))
	     ) CT.empty (Ctype.get_fields rd))
    | _ -> 
	raise Exit

and typ_from_ml_constr env p args =
  try 
    let (_,args',n) = 
      List.find (fun (p',_,_) -> Path.same p p') env.constrs in
    List.iter2 (Ctype.unify !solve_env) args args';
    n 
  with Ctype.Unify _ -> raise Exit | Not_found -> 
    let n = CT.make () in
    let env = { env with constrs = (p,args,n) :: env.constrs } in
    let decl = 
      try Env.find_type p !solve_env
      with Not_found -> raise Exit in
    let inst t =
      let ps,t = Ctype.instance_parameterized_type decl.type_params t  in
      List.iter2 (Ctype.unify !solve_env) ps args;
      t in
    let t : CT.t = 
      match decl.type_kind, decl.type_manifest with
	  (* 	       | Type_abstract, Some t -> typ_from_ml env (inst t) *)
	| Type_abstract, None when Path.same p Predef.path_float ->
	    CT.abstract (CT.Abstract.atom (Path.name p))
	| Type_variant (cstrs,_), _ ->
	    List.fold_left
	      (fun accu (c,p) ->
		 let p = List.map inst p in
		 let p = List.map (typ_from_ml env) p in
		 let t = ml_constr c p in
		 CT.cup accu t)
	      CT.empty cstrs
	| Type_record (fields,_,_), _ ->
	    let fields =
	      LabelMap.from_list (fun _ _ -> assert false)
		(
		  List.map
		    (fun (lab,_,t) ->
		       let t = typ_from_ml env (inst t) in
		       (LabelPool.mk (Ns.empty,U.mk_latin1 lab), t)
		    ) fields) in
	    CT.record_fields (false,fields)
	| _ -> raise Exit in
    CT.define n t;
    n
      
let typ_from_ml loc t = 
  try CT.descr (typ_from_ml { loc = loc; seen =  []; constrs = [] }  t)
  with Exit -> error loc (CannotTranslateML t)

(*********************)


let parse_ns env loc pr =
  try Env.find_namespace pr env
  with Cduce_types.Ns.UnknownPrefix ns ->
    error loc (UnboundNamespacePrefix (U.to_string ns))

let parse_atom env loc (pr,l) =
  let ns = parse_ns env loc pr in
  Cduce_types.Atoms.V.mk ns l
 
let transl_label env loc (pr,l) =
  let ns = if (pr = "") then Ns.empty else parse_ns env loc pr in
  Cduce_types.Ident.LabelPool.mk (ns,l)

let transl_record (env : Env.t) loc f r :
    'a Cduce_types.Ident.label_map =
  let r = List.map (fun (q,x) -> (transl_label env loc q, f x)) r in
  LabelMap.from_list (fun _ _ ->  error loc DuplicatedLabel) r

let rec transl_const env c = match c.pextcst_desc with
  | Pextcst_pair (c1,c2) -> CT.Pair (transl_const env c1, transl_const env c2)
  | Pextcst_xml (c1,c2,c3) -> 
      CT.Xml (transl_const env c1, 
	      CT.Pair (transl_const env c2,transl_const env c3))
  | Pextcst_record f -> 
      CT.Record (transl_record env c.pextcst_loc (transl_const env) f)
  | Pextcst_atom q -> CT.Atom (parse_atom env c.pextcst_loc q)
  | Pextcst_int i -> CT.Integer i
  | Pextcst_char c -> CT.Char c
  | Pextcst_string s -> CT.String (U.start_index s,U.end_index s,s,SEQ.nil_cst)
  | Pextcst_intern c -> c

(* Eliminate Recursion, propagate Sequence Capture Variables *)

type derecurs_slot = {
  ploc : Location.t;
  pid  : int;
  mutable ploop : bool;
  mutable pdescr : derecurs;
} and derecurs =
  | PDummy
  | PAlias of derecurs_slot
  | PType of CT.t
  | POr of derecurs * derecurs
  | PAnd of derecurs * derecurs
  | PDiff of derecurs * derecurs
  | PTimes of derecurs * derecurs
  | PXml of derecurs * derecurs
  | PArrow of derecurs * derecurs
  | POptional of derecurs
  | PRecord of bool * (derecurs * derecurs option) Cduce_types.Ident.label_map
  | PCapture of Cduce_types.Ident.id
  | PConstant of Cduce_types.Ident.id * CT.const
  | PRegexp of derecurs_regexp
and derecurs_regexp =
  | PEpsilon
  | PElem of derecurs
  | PGuard of derecurs
  | PSeq of derecurs_regexp * derecurs_regexp
  | PAlt of derecurs_regexp * derecurs_regexp
  | PStar of derecurs_regexp
  | PWeakStar of derecurs_regexp

type descr = 
  | IDummy
  | IType of CT.descr
  | IOr of descr * descr
  | IAnd of descr * descr
  | IDiff of descr * descr
  | ITimes of slot * slot
  | IXml of slot * slot
  | IArrow of slot * slot
  | IOptional of descr
  | IRecord of bool * (slot * descr option) Cduce_types.Ident.label_map
  | ICapture of Cduce_types.Ident.id
  | IConstant of Cduce_types.Ident.id * CT.const
and slot = {
  mutable fv : Cduce_types.Ident.fv option;
  mutable hash : int option;
  mutable rank1: int; mutable rank2: int;
  mutable gen1 : int; mutable gen2: int;
  mutable d    : descr;
}
    

let counter = ref 0
let mk_derecurs_slot loc = 
  incr counter; 
  { ploop = false; ploc = loc; pid = !counter; pdescr = PDummy }
	  
let mk_slot () = 
  { d=IDummy; fv=None; hash=None; rank1=0; rank2=0; gen1=0; gen2=0 } 


(* This environment is used in phase (1) to eliminate recursion *)
type penv = {
  penv_tenv : Env.t;
  penv_derec : derecurs_slot CEnv.t;
}


let penv tenv = { penv_tenv = tenv; penv_derec = CEnv.empty }

let rec hash_derecurs = function
  | PDummy -> assert false
  | PAlias s -> 
      s.pid
  | PType t -> 
      1 + 17 * (CT.hash t)
  | POr (p1,p2) -> 
      2 + 17 * (hash_derecurs p1) + 257 * (hash_derecurs p2)
  | PAnd (p1,p2) -> 
      3 + 17 * (hash_derecurs p1) + 257 * (hash_derecurs p2)
  | PDiff (p1,p2) -> 
      4 + 17 * (hash_derecurs p1) + 257 * (hash_derecurs p2)
  | PTimes (p1,p2) -> 
      5 + 17 * (hash_derecurs p1) + 257 * (hash_derecurs p2)
  | PXml (p1,p2) -> 
      6 + 17 * (hash_derecurs p1) + 257 * (hash_derecurs p2)
  | PArrow (p1,p2) -> 
      7 + 17 * (hash_derecurs p1) + 257 * (hash_derecurs p2)
  | POptional p -> 
      8 + 17 * (hash_derecurs p)
  | PRecord (o,r) -> 
      (if o then 9 else 10) + 
	17 * (LabelMap.hash hash_derecurs_field r)
  | PCapture x -> 
      11 + 17 * (Cduce_types.Ident.Id.hash x)
  | PConstant (x,c) -> 
      12 + 17 * (Cduce_types.Ident.Id.hash x) + 257 * (CT.Const.hash c)
  | PRegexp p -> 
      13 + 17 * (hash_derecurs_regexp p)
and hash_derecurs_field = function
  | (p, Some e) -> 1 + 17 * hash_derecurs p + 257 * hash_derecurs e
  | (p, None) -> 2 + 17 * hash_derecurs p
and hash_derecurs_regexp = function
  | PEpsilon -> 
      1
  | PElem p -> 
      2 + 17 * (hash_derecurs p)
  | PSeq (p1,p2) -> 
      3 + 17 * (hash_derecurs_regexp p1) + 257 * (hash_derecurs_regexp p2)
  | PAlt (p1,p2) -> 
      4 + 17 * (hash_derecurs_regexp p1) + 257 * (hash_derecurs_regexp p2)
  | PStar p -> 
      5 + 17 * (hash_derecurs_regexp p)
  | PWeakStar p -> 
      6 + 17 * (hash_derecurs_regexp p)
  | PGuard p ->
      7 + 17 * (hash_derecurs p)

let rec equal_derecurs p1 p2 = (p1 == p2) || match p1,p2 with
  | PAlias s1, PAlias s2 -> 
      s1 == s2
  | PType t1, PType t2 -> 
      CT.equal t1 t2
  | POr (p1,q1), POr (p2,q2)
  | PAnd (p1,q1), PAnd (p2,q2)
  | PDiff (p1,q1), PDiff (p2,q2)
  | PTimes (p1,q1), PTimes (p2,q2)
  | PXml (p1,q1), PXml (p2,q2)
  | PArrow (p1,q1), PArrow (p2,q2) -> 
      (equal_derecurs p1 p2) && (equal_derecurs q1 q2)
  | POptional p1, POptional p2 -> 
      equal_derecurs p1 p2
  | PRecord (o1,r1), PRecord (o2,r2) -> 
      (o1 == o2) && 
	(LabelMap.equal equal_derecurs_field r1 r2)
  | PCapture x1, PCapture x2 -> 
      Cduce_types.Ident.Id.equal x1 x2
  | PConstant (x1,c1), PConstant (x2,c2) -> 
      (Cduce_types.Ident.Id.equal x1 x2) && (CT.Const.equal c1 c2)
  | PRegexp p1, PRegexp p2 -> 
      equal_derecurs_regexp p1 p2
  | _ -> false
and equal_derecurs_field r1 r2 = match (r1,r2) with
  | (p1,None),(p2,None) -> equal_derecurs p1 p2
  | (p1, Some e1), (p2, Some e2) -> equal_derecurs p1 p2 && equal_derecurs e1 e2
  | _ -> false
and equal_derecurs_regexp r1 r2 = match r1,r2 with
  | PEpsilon, PEpsilon -> 
      true
  | PElem p1, PElem p2 -> 
      equal_derecurs p1 p2
  | PGuard p1, PGuard p2 ->
      equal_derecurs p1 p2
  | PSeq (p1,q1), PSeq (p2,q2) 
  | PAlt (p1,q1), PAlt (p2,q2) -> 
      (equal_derecurs_regexp p1 p2) && (equal_derecurs_regexp q1 q2)
  | PStar p1, PStar p2
  | PWeakStar p1, PWeakStar p2 -> 
      equal_derecurs_regexp p1 p2
  | _ -> false

module DerecursTable = Hashtbl.Make(
  struct 
    type t = derecurs 
    let hash = hash_derecurs
    let equal = equal_derecurs
  end
)

let gen = ref 0
let rank = ref 0
	     
let rec hash_descr = function
  | IDummy -> assert false
  | IType x -> CT.hash x
  | IOr (d1,d2) -> 1 + 17 * (hash_descr d1) + 257 * (hash_descr d2)
  | IAnd (d1,d2) -> 2 + 17 * (hash_descr d1) + 257 * (hash_descr d2)
  | IDiff (d1,d2) -> 3 + 17 * (hash_descr d1) + 257 * (hash_descr d2)
  | IOptional d -> 4 + 17 * (hash_descr d)
  | ITimes (s1,s2) -> 5 + 17 * (hash_slot s1) + 257 * (hash_slot s2)
  | IXml (s1,s2) -> 6 + 17 * (hash_slot s1) + 257 * (hash_slot s2)
  | IArrow (s1,s2) -> 7 + 17 * (hash_slot s1) + 257 * (hash_slot s2)
  | IRecord (o,r) -> 
      (if o then 8 else 9) + 17 * 
	(LabelMap.hash hash_descr_field r)
  | ICapture x -> 10 + 17 * (Cduce_types.Ident.Id.hash x)
  | IConstant (x,y) -> 
      11 + 17 * (Cduce_types.Ident.Id.hash x) + 257 * (CT.Const.hash y)
and hash_descr_field = function
  | (d, Some e) -> 1 + 17 * hash_slot d + 257 * hash_descr e
  | (d, None) -> 2 + 17 * hash_slot d
and hash_slot s =
  if s.gen1 = !gen then 13 * s.rank1
  else (
    incr rank;
    s.rank1 <- !rank; s.gen1 <- !gen;
    hash_descr s.d
  )
    
let rec equal_descr d1 d2 = 
  match (d1,d2) with
  | IType x1, IType x2 -> CT.equal x1 x2
  | IOr (x1,y1), IOr (x2,y2) 
  | IAnd (x1,y1), IAnd (x2,y2) 
  | IDiff (x1,y1), IDiff (x2,y2) -> (equal_descr x1 x2) && (equal_descr y1 y2)
  | IOptional x1, IOptional x2 -> equal_descr x1 x2
  | ITimes (x1,y1), ITimes (x2,y2) 
  | IXml (x1,y1), IXml (x2,y2) 
  | IArrow (x1,y1), IArrow (x2,y2) -> (equal_slot x1 x2) && (equal_slot y1 y2)
  | IRecord (o1,r1), IRecord (o2,r2) -> 
      (o1 = o2) && (LabelMap.equal equal_descr_field r1 r2)
  | ICapture x1, ICapture x2 -> Cduce_types.Ident.Id.equal x1 x2
  | IConstant (x1,y1), IConstant (x2,y2) -> 
      (Cduce_types.Ident.Id.equal x1 x2) && (CT.Const.equal y1 y2)
  | _ -> false
and equal_descr_field d1 d2 = match (d1,d2) with
  | (d1,None),(d2,None) -> equal_slot d1 d2
  | (d1, Some e1), (d2, Some e2) -> equal_slot d1 d2 && equal_descr e1 e2
  | _ -> false
and equal_slot s1 s2 =
  ((s1.gen1 = !gen) && (s2.gen2 = !gen) && (s1.rank1 = s2.rank2))
  ||
  ((s1.gen1 <> !gen) && (s2.gen2 <> !gen) && (
     incr rank;
     s1.rank1 <- !rank; s1.gen1 <- !gen;
     s2.rank2 <- !rank; s2.gen2 <- !gen;
     equal_descr s1.d s2.d
   ))
  
module SlotTable = Hashtbl.Make(
  struct
    type t = slot
	
    let hash s =
      match s.hash with
	| Some h -> h
	| None ->
	    incr gen; rank := 0; 
	    let h = hash_slot s in
	    s.hash <- Some h;
	    h
	      
    let equal s1 s2 = 
      (s1 == s2) || 
      (incr gen; rank := 0; 
       let e = equal_slot s1 s2 in
       (*     if e then Printf.eprintf "Recursive hash-consing: Equal\n";  *)
       e)
  end)


let pempty = PType CT.empty

let por p1 p2 =
  if p1 == pempty then p2 else
    if p2 == pempty then p1 else
      POr (p1,p2)

let pand p1 p2 =
  if (p1 == pempty) || (p2 == pempty) then pempty else PAnd (p1,p2)

let rec remove_regexp r q = match r with
  | PEpsilon ->
      q
  | PElem p ->
      PTimes (p, q)
  | PGuard p ->
      pand p q
  | PSeq (r1,r2) ->
      remove_regexp r1 (remove_regexp r2 q)
  | PAlt (r1,r2) ->
      por (remove_regexp r1 q) (remove_regexp r2 q)
  | PStar r ->
      let x = mk_derecurs_slot noloc in
      let res = POr (PAlias x, q) in
      x.pdescr <- remove_regexp2 r res pempty;
      res
  | PWeakStar r ->
      let x = mk_derecurs_slot noloc in
      let res = POr (q, PAlias x) in
      x.pdescr <- remove_regexp2 r res pempty;
      res

and remove_regexp2 r q_nonempty q_empty =
  if q_nonempty == q_empty then remove_regexp r q_empty
  else match r with
    | PEpsilon ->
        q_empty
    | PElem p ->
        PTimes (p, q_nonempty)
    | PGuard p ->
	pand p q_empty
    | PSeq (r1,r2) ->
        remove_regexp2 r1
        (remove_regexp2 r2 q_nonempty q_nonempty)
        (remove_regexp2 r2 q_nonempty q_empty)
    | PAlt (r1,r2) ->
        por
        (remove_regexp2 r1 q_nonempty q_empty)
        (remove_regexp2 r2 q_nonempty q_empty)
    | PStar r ->
        let x = mk_derecurs_slot noloc in
        x.pdescr <- remove_regexp2 r (POr (PAlias x, q_nonempty)) pempty;
        por (PAlias x) q_empty
    | PWeakStar r ->
        let x = mk_derecurs_slot noloc in
        x.pdescr <- remove_regexp2 r (POr (q_nonempty, PAlias x)) pempty;
        por q_empty (PAlias x)

let cst_nil = CT.Atom SEQ.nil_atom
let capture_all vars p =
  IdSet.fold (fun p x -> PAnd (p, PCapture x)) p vars
let termin b vars p = 
  if b then p 
  else IdSet.fold 
    (fun p x -> PSeq (p, PGuard (PConstant (x,cst_nil)))) p vars

let pcdata =
  Pext_star (Pext_elem { pext_loc = Location.none;
			 pext_desc = Pext_cst CT.Char.any })

let rec derecurs env p = match p.pext_desc with
  | Pext_name v -> derecurs_var env p.pext_loc v
  | Pext_recurs (p,b) -> derecurs (derecurs_def env b) p
  | Pext_cst t -> PType t
  | Pext_ns ns -> PType 
      (CT.atom (Cduce_types.Atoms.any_in_ns (parse_ns env.penv_tenv p.pext_loc ns)))
  | Pext_or (p1,p2) -> POr (derecurs env p1, derecurs env p2)
  | Pext_and (p1,p2) -> PAnd (derecurs env p1, derecurs env p2)
  | Pext_diff (p1,p2) -> PDiff (derecurs env p1, derecurs env p2)
  | Pext_prod (p1,p2) -> PTimes (derecurs env p1, derecurs env p2)
  | Pext_xml (p1,p2) -> PXml (derecurs env p1, derecurs env p2)
  | Pext_arrow (p1,p2) -> PArrow (derecurs env p1, derecurs env p2) 
  | Pext_optional p -> POptional (derecurs env p)
  | Pext_record (o,r) -> 
      let aux = function
	| (p,Some e) -> (derecurs env p, Some (derecurs env e))
	| (p,None) -> derecurs env p, None in
      PRecord (o, transl_record env.penv_tenv p.pext_loc aux r)
  | Pext_bind (x,c) -> 
      PConstant (id x,transl_const env.penv_tenv c)
  | Pext_constant c -> PType (CT.constant (transl_const env.penv_tenv c))
  | Pext_regexp r ->
      let r,_ = derecurs_regexp IdSet.empty false 
	IdSet.empty true env r in
      PRegexp r
	(* Note: computing remove_regexp here is slower (because
	   of caching ?) *)
  | Pext_from_ml t ->
      (match !transl_simple_type with
	 | None -> assert false
	 | Some f ->
	     PType (
	       typ_from_ml p.pext_loc
		 (f env.penv_tenv true t)))

and derecurs_regexp vars b rvars f env = function
(* - vars: seq variables to be propagated top-down and added
     to each captured element
   - b: below a star ?
   - rvars: seq variables that appear on the right of the regexp
   - f: tail position

  returns the set of seq variable of the regexp minus rvars
  (they have already been terminated if not below a star)
*)
  | Pext_epsilon -> 
      PEpsilon, IdSet.empty
  | Pext_elem { pext_desc = Pext_name (Longident.Lident "PCDATA") } ->
      derecurs_regexp vars b rvars f env pcdata
  | Pext_elem p -> 
      PElem (capture_all vars (derecurs env p)), IdSet.empty
  | Pext_guard p ->
      PGuard (derecurs env p), IdSet.empty
  | Pext_seq (p1,p2) -> 
      let (p2,v2) = derecurs_regexp vars b rvars f env p2 in
      let (p1,v1) = derecurs_regexp vars b (IdSet.cup rvars v2) false env p1 in
      PSeq (p1,p2), IdSet.cup v1 v2
  | Pext_alt (p1,p2) -> 
      let (p1,v1) = derecurs_regexp vars b rvars f env p1
      and (p2,v2) = derecurs_regexp vars b rvars f env p2 in
      PAlt (termin b (IdSet.diff v2 v1) p1, termin b (IdSet.diff v1 v2) p2),
      IdSet.cup v1 v2
  | Pext_star p -> 
      let (p,v) = derecurs_regexp vars true rvars false env p in
      termin b v (PStar p), v
  | Pext_weakstar p -> 
      let (p,v) = derecurs_regexp vars true rvars false env p in
      termin b v (PWeakStar p), v
  | Pext_capture (x,p) -> 
      let x = id x in
      let vars = if f then vars else IdSet.add x vars in
      let after = IdSet.mem rvars x in
      let rvars = IdSet.add x rvars in
      let (p,v) = derecurs_regexp vars b rvars false env p in
      (if f 
       then PSeq (PGuard (PCapture x), p) 
       else termin (after || b) (IdSet.singleton x) p), 
      (if after then v else IdSet.add x v)


and derecurs_var env loc v =
  match v with
    | Longident.Lident x ->
	let id = id x in
	(try
	   try PType (CEnv.find id built_ins)
	   with Not_found ->
	     try PAlias (CEnv.find id env.penv_derec)
	     with Not_found -> derecurs_var' env loc v
	 with Not_found -> PCapture id)
    | _ -> 
	try derecurs_var' env loc v
	with Not_found -> error loc NotXmlType

and derecurs_var' env loc v = 
  let (p,d) = Env.lookup_type v env.penv_tenv in
  match d.Types.type_manifest with
     | Some ty ->
	 (match (repr (expand_head env.penv_tenv ty)).desc with
	    | Types.Text { Types.ext_const = Some t } -> PType t
	    | _ -> error loc NotXmlType)
     | _ -> error loc NotXmlType
	
and derecurs_def env b =
  let b = List.map 
    (fun (v,p) -> (id v,p,mk_derecurs_slot p.pext_loc)) b in
  let n = 
    List.fold_left (fun env (v,p,s) -> CEnv.add v s env) env.penv_derec b in
  let env = { env with penv_derec = n } in
  List.iter (fun (v,p,s) -> s.pdescr <- derecurs env p) b;
  env


let rec fv_slot s =
  match s.fv with
    | Some x -> x
    | None ->
	if s.gen1 = !gen then IdSet.empty 
	else (s.gen1 <- !gen; fv_descr s.d)
and fv_descr = function
  | IDummy -> assert false
  | IType _ -> IdSet.empty
  | IOr (d1,d2)
  | IAnd (d1,d2)  
  | IDiff (d1,d2) -> IdSet.cup (fv_descr d1) (fv_descr d2)
  | IOptional d -> fv_descr d
  | ITimes (s1,s2)  
  | IXml (s1,s2)  
  | IArrow (s1,s2) -> IdSet.cup (fv_slot s1) (fv_slot s2)
  | IRecord (o,r) -> 
      List.fold_left IdSet.cup IdSet.empty 
	(LabelMap.map_to_list fv_field r)
  | ICapture x | IConstant (x,_) -> IdSet.singleton x
and fv_field = function
  | (d,Some e) -> IdSet.cup (fv_slot d) (fv_descr e)
  | (d,None) -> fv_slot d


let compute_fv s =
  match s.fv with
    | Some x -> ()
    | None ->
	incr gen;
	let x = fv_slot s in
	s.fv <- Some x
	  
let check_no_capture loc s =
  match IdSet.pick s with
    | Some x -> error loc (CaptureNotAllowed (Cduce_types.Ident.to_string x))
    | None -> ()
    
let compile_slot_hash = DerecursTable.create 67
let compile_hash = DerecursTable.create 67

let todo_defs = ref []
let todo_fv = ref []

let rec compile p =
  try DerecursTable.find compile_hash p
  with Not_found ->
    let c = real_compile p in
    DerecursTable.replace compile_hash p c;
    c
and real_compile = function
  | PDummy -> assert false
  | PAlias v ->
      if v.ploop then error v.ploc UnguardedRecursion;
      v.ploop <- true;
      let r = compile v.pdescr in
      v.ploop <- false;
      r
  | PType t -> IType t
  | POr (t1,t2) -> IOr (compile t1, compile t2)
  | PAnd (t1,t2) -> IAnd (compile t1, compile t2)
  | PDiff (t1,t2) -> IDiff (compile t1, compile t2)
  | PTimes (t1,t2) -> ITimes (compile_slot t1, compile_slot t2)
  | PXml (t1,t2) -> IXml (compile_slot t1, compile_slot t2)
  | PArrow (t1,t2) -> IArrow (compile_slot t1, compile_slot t2)
  | POptional t -> IOptional (compile t)
  | PRecord (o,r) ->  IRecord (o, LabelMap.map compile_field r)
  | PConstant (x,v) -> IConstant (x,v)
  | PCapture x -> ICapture x
  | PRegexp r -> compile (remove_regexp r (PType SEQ.nil_type))

and compile_field = function
  | (p, Some e) -> (compile_slot p, Some (compile e))
  | (p, None) -> (compile_slot p, None)

and compile_slot p =
  try DerecursTable.find compile_slot_hash p
  with Not_found ->
    let s = mk_slot () in
    todo_defs := (s,p) :: !todo_defs;
    todo_fv := s :: !todo_fv;
    DerecursTable.add compile_slot_hash p s;
    s

      
let rec flush_defs () = 
  match !todo_defs with
    | [] -> 
	List.iter compute_fv !todo_fv;
	todo_fv := [];
    | (s,p)::t -> 
	todo_defs := t; 
	s.d <- compile p; 
	flush_defs ()
	
let typ_nodes = SlotTable.create 67
let pat_nodes = SlotTable.create 67
		  
let rec typ = function
  | IType t -> t
  | IOr (s1,s2) -> CT.cup (typ s1) (typ s2)
  | IAnd (s1,s2) ->  CT.cap (typ s1) (typ s2)
  | IDiff (s1,s2) -> CT.diff (typ s1) (typ s2)
  | ITimes (s1,s2) -> CT.times (typ_node s1) (typ_node s2)
  | IXml (s1,s2) -> CT.xml (typ_node s1) (typ_node s2)
  | IArrow (s1,s2) -> CT.arrow (typ_node s1) (typ_node s2)
  | IOptional s -> CT.Record.or_absent (typ s)
  | IRecord (o,r) ->  CT.record_fields (o, LabelMap.map typ_field r)
  | IDummy | ICapture _ | IConstant (_,_) -> assert false
      
and typ_field = function
  | (s, None) -> typ_node s
  | (s, Some _) -> 
      raise (Patterns.Error "Or-else clauses are not allowed in types")

and typ_node s : CT.Node.t =
  try SlotTable.find typ_nodes s
  with Not_found ->
    let x = CT.make () in
    SlotTable.add typ_nodes s x;
    CT.define x (typ s.d);
    x
      
let rec pat d : Patterns.descr =
  if IdSet.is_empty (fv_descr d)
  then Patterns.constr (typ d)
  else pat_aux d
    
and pat_aux = function
  | IDummy -> assert false
  | IOr (s1,s2) -> Patterns.cup (pat s1) (pat s2)
  | IAnd (s1,s2) -> Patterns.cap (pat s1) (pat s2)
  | IDiff (s1,s2) when IdSet.is_empty (fv_descr s2) ->
      let s2 = CT.neg (typ s2) in
      Patterns.cap (pat s1) (Patterns.constr s2)
  | IDiff _ ->
      raise (Patterns.Error "Differences are not allowed in patterns")
  | ITimes (s1,s2) -> Patterns.times (pat_node s1) (pat_node s2)
  | IXml (s1,s2) -> Patterns.xml (pat_node s1) (pat_node s2)
  | IOptional _ -> 
      raise (Patterns.Error "Optional fields are not allowed in record patterns")
  | IRecord (o,r) ->
      let pats = ref [] in
      let aux l = function
	| (s,None) ->
	    if IdSet.is_empty (fv_slot s) then typ_node s
	    else
	      ( pats := Patterns.record l (pat_node s) :: !pats;
		CT.any_node )
	| (s,Some e) ->
	    if IdSet.is_empty (fv_slot s) then
	      raise (Patterns.Error "Or-else clauses are not allowed in types")
	    else
	      ( pats := Patterns.cup 
		  (Patterns.record l (pat_node s))
		  (pat e) :: !pats;
		CT.Record.any_or_absent_node )
      in
      let constr = CT.record_fields (o,LabelMap.mapi aux r) in
      List.fold_left Patterns.cap (Patterns.constr constr) !pats
	(* TODO: can avoid constr when o=true, and all fields have fv *)
  | ICapture x -> Patterns.capture x
  | IConstant (x,c) -> Patterns.constant x c
  | IArrow _ ->
      raise (Patterns.Error "Arrows are not allowed in patterns")
  | IType _ -> assert false
      
and pat_node s : Patterns.node =
  try SlotTable.find pat_nodes s
  with Not_found ->
    let x = Patterns.make (fv_slot s) in
    try
      SlotTable.add pat_nodes s x;
      Patterns.define x (pat s.d);
      x
    with exn -> SlotTable.remove pat_nodes s; raise exn
      (* For the toplevel ... *)


let do_typ loc r = 
  let s = compile_slot r in
  flush_defs ();
  check_no_capture loc (fv_slot s);
  typ_node s
   
let transl_ext_type_node env p =
  solve_env := env;
  do_typ p.pext_loc (derecurs (penv env) p)
    
let transl_ext_pat env p = 
  solve_env := env;
  let s = compile_slot (derecurs (penv env) p) in
  flush_defs ();
  pat_node s

let transl_ext_type env p =
  let t = CT.descr (transl_ext_type_node env p) in
  Btype.newextvar 
    { Types.ext_const = Some t; Types.ext_atoms = []; Types.ext_lb = [] }

let transl_type_decl env b =
  solve_env := env;
  (* Mmmh, if t is already defined, type t = {{ {: t :} }} will be
     accepted *)

  let penv = derecurs_def (penv env) b in
  let decl = List.map (fun (_,p) -> (p,compile (derecurs penv p))) b in
  flush_defs ();
  let ts = List.map 
    (fun (p,s) -> 
       check_no_capture p.pext_loc (fv_descr s);
       let t = typ s in
       if (p.pext_loc <> noloc) && (CT.is_empty t) then 
	 error p.pext_loc EmptyType;
       t) decl in
  List.iter2 (fun (v,_) t -> CT.Print.register_global CT.CompUnit.pervasives (Ns.empty, U.mk_latin1 v) t) 
    b ts;
  ts
    

let transl_branches env b =
  let t,l = 
    List.fold_left 
      (fun (acc,bl) (p,e) -> 
	 let p = transl_ext_pat env p in
	 let a = CT.descr (Patterns.accept p) in
	 (CT.cup acc a),(CT.diff a acc,p,e)::bl
      ) (CT.empty,[]) b
  in
  t, List.rev l






(*****************************************************************************)
(* TODO: return a lazy type to avoid computing it when not necessary *)


let exptypes = ref []

let register_ext_annot env s e =
  match (repr (expand_head env e.Typedtree.exp_type)).desc with
    | Tvar -> exptypes := (s,e) :: !exptypes
    | Text _ -> s.Parsetree.pexp_ext <- true
    | _ -> ()

let flush_ext_annot env =
  List.iter 
    (fun (s,e) ->
       match (repr (expand_head env e.Typedtree.exp_type)).desc with
	 | Text _ -> s.Parsetree.pexp_ext <- true
	 | _ -> ()
    ) !exptypes;
  exptypes := []




let anyext_var = newextvar anyext

let force_ext env loc t =
  let a = newextvar anyext in
  unify env a t

let pair_subtype loc =
  let a = newextvar anyext in
  let b = newextvar { anyext with ext_lb = [loc,a] } in
  (a,b)
  
let ext_subtype env loc t =
  let (a,b) = pair_subtype loc in
  unify env t b;
  a
  
let ext_supertype env loc t =   
  let (a,b) = pair_subtype loc in
  unify env t a;
  b

let ext_cst env loc c =
  if !extmode then anyext_var
  else newextvar { ext_const = Some (CT.constant c); 
		   ext_lb = []; ext_atoms = [] }

let atom loc f =
  newextvar { anyext with ext_atoms = 
      [ loc, { ext_atom_v = Atom_start; ext_atom_def = f } ] }

let ext_ub env loc ub e =
  if !extmode then anyext_var 
  else atom loc (fun _ -> compute_var loc e ub)


let ext_from_ml env loc t =
  if !extmode then anyext_var 
  else atom loc (fun _ -> typ_from_ml loc t)

let ext_to_ml env loc ml ext =
  if !extmode then ()
  else ignore (atom loc (fun _ -> compute_var loc ext (typ_from_ml loc ml)))

let ext_int2 f env loc e1 e2 =
  if !extmode then anyext_var 
  else atom loc 
    (fun ub ->
       let t1 = compute_var loc e1 CT.Int.any in
       let t2 = compute_var loc e2 CT.Int.any in
       CT.interval
	 (f (CT.Int.get t1) (CT.Int.get t2)))

let ext_add env loc e1 e2 =
  if !extmode then anyext_var 
  else atom loc 
    (fun ub ->
       let t1 = compute_var loc e1 (CT.cup CT.Int.any CT.Record.any) in
       if CT.subtype t1 CT.Int.any
       then
	 let t2 = compute_var loc e2 CT.Int.any in
	 CT.interval 
	   (Cduce_types.Intervals.add (CT.Int.get t1) (CT.Int.get t2))
       else if CT.subtype t1 CT.Record.any
       then
	 let t2 = compute_var loc e2 CT.Record.any in
	 CT.Record.merge t1 t2
       else
	 error loc (AddMix t1))


let add_values =
  List.fold_left
    (fun env (x,v) -> Env.add_value x { val_type = v; val_kind = Val_reg } env)

let ext_branch env loc t p a =
  let module P = Cduce_types.Patterns in
  let fv = P.fv p in
  let ids = 
    List.map (fun x -> x, Ident.create (Cduce_types.Ident.to_string x)) fv in
  let vars = 
    if !extmode then List.map (fun (_,x) -> x, anyext_var) ids
    else
      let z = lazy (
	let ta = compute_var loc a CT.any in
	P.filter (CT.cap t ta) p 
      ) in
      List.map
	(fun (x,id) ->
	   id, 
	   atom loc (fun _ -> CT.descr (Cduce_types.Ident.IdMap.assoc x (Lazy.force z))))
	ids
  in
  ids, add_values env vars

let ext_pair env loc e1 e2 =
  if !extmode then anyext_var
  else atom loc
    (fun ub ->
       let t = CT.Product.get ub in
       let t1 = compute_var loc e1 (CT.Product.pi1 t) in
       let t2 = compute_var loc e2 (CT.Product.pi2 t) in
       CT.times (CT.cons t1) (CT.cons t2))

let ext_record (env : Env.t) loc fl =
  if !extmode then anyext_var
  else atom loc
    (fun ub ->
       let fl = 
	 LabelMap.map 
	   (fun e -> CT.cons (compute_var loc e CT.any))
	   fl in
       CT.record_fields (false,fl))

let ext_concat env loc e1 e2 =
  if !extmode then anyext_var
  else atom loc
    (fun ub ->
       let t1 = compute_var loc e1 SEQ.any in
       let t2 = compute_var loc e2 SEQ.any in
       SEQ.concat t1 t2)

let ext_xml env loc e1 e2 e3 =
  if !extmode then anyext_var
  else atom loc
    (fun ub ->
       let t1 = compute_var loc e1 CT.any in
       let t2 = compute_var loc e2 CT.any in
       let t3 = compute_var loc e3 CT.any in
       CT.xml
	 (CT.cons t1) (CT.cons (CT.times (CT.cons t2) (CT.cons t3))))

let ext_apply env loc e1 e2 =
  if !extmode then anyext_var
  else atom loc
    (fun ub ->
       let module A = CT.Arrow in
       let t1 = compute_var loc e1 A.any in
       let t1 = A.get t1 in
       let dom = A.domain t1 in
       let t2 = compute_var loc e2 (A.domain t1)  in
       A.apply t1 t2)

let ext_map env loc e0 e1 brs =
  if !extmode then (unify env e1 anyext_var; anyext_var)
  else 
    let z = lazy (
      let t0 = compute_var loc e0 CT.any in
      let tl,f = SEQ.map_mono t0 in
      let u = List.fold_left CT.cup CT.empty tl in
      (u,(tl,f))
    ) in
    unify env e1 (atom loc (fun ub -> fst (Lazy.force z)));
    atom loc 
      (fun ub ->
	 let (tl,f) = snd (Lazy.force z) in
	 let aux t = 
	   List.fold_left
	     (fun accu (s,e) ->
		if (CT.disjoint t s) then accu
		else CT.cup accu (compute_var loc e SEQ.any))
	     CT.empty brs
	 in
	 SEQ.flatten (f (List.map aux tl))
      )

let ext_xmap env loc domain e0 e1 brs =
  if !extmode then (unify env e1 anyext_var; anyext_var)
  else 
    let z = lazy (
      let t0 = compute_var loc e0 CT.any in
      let u, tl,f = SEQ.map_tree_mono domain t0 in
      (u,(tl,f))
    ) in
    unify env e1 (atom loc (fun ub -> fst (Lazy.force z)));
    atom loc 
      (fun ub ->
	 let (tl,f) = snd (Lazy.force z) in
	 let aux t = 
	   List.fold_left
	     (fun accu (s,e) ->
		if (CT.disjoint t s) then accu
		else CT.cup accu (compute_var loc e SEQ.any))
	     CT.empty brs
	 in
	 f (List.map aux tl)
      )
     
let ext_removefield env loc e l =
  if !extmode then anyext_var 
  else atom loc 
    (fun ub ->
       let t = compute_var loc e CT.Record.any in
       CT.Record.remove_field t l)

let ext_op env loc op tl = match op,tl with
  | "concat",[t1;t2] -> ext_concat env loc t1 t2
  | "add",[t1;t2] -> ext_add env loc t1 t2
  | "mul",[t1;t2] -> ext_int2 Cduce_types.Intervals.mul env loc t1 t2
  | "sub",[t1;t2] -> ext_int2 Cduce_types.Intervals.sub env loc t1 t2
  | "div",[t1;t2] -> ext_int2 Cduce_types.Intervals.div env loc t1 t2
  | "modulo",[t1;t2] -> ext_int2 Cduce_types.Intervals.modulo env loc t1 t2
  | "pair",[t1;t2] -> ext_pair env loc t1 t2
  | "xml",[t1;t2;t3] -> ext_xml env loc t1 t2 t3
  | "apply",[t1;t2] -> ext_apply env loc t1 t2
  | _ -> assert false

  



open Typedtree

let type_expect ?in_function env sexp ty_expected =
  if !extmode then 
    let e = !real_type_expect ?in_function env sexp ty_expected in
    register_ext_annot env sexp e; 
    e
  else
    if sexp.pexp_ext 
    then 
      let t = ext_subtype env sexp.pexp_loc ty_expected in
      let e = !real_type_expect ?in_function env sexp t in
      { e with exp_type = ty_expected }
    else
      !real_type_expect ?in_function env sexp ty_expected


let type_ext env e =
  type_expect env e 
    (newextvar { ext_const = None; ext_lb = []; ext_atoms = [] })

let type_expression env loc = function
  | Pextexp_cst cst ->
      let cst = transl_const env cst in
      Textexp_cst cst, 
      ext_cst env loc cst
  | Pextexp_match (e0,bl) ->
      let e0 = type_ext env e0 in
      let res = newvar () in

      let (acc,bl) = transl_branches env bl in
      let t0 = ext_ub env e0.exp_loc acc e0.exp_type in
      let brs = List.map
	(fun (t,p,e) ->
	   let binds,newenv = ext_branch env e.pexp_loc t p t0 in
	   let e = type_expect newenv e res in
	   (p,binds,e)) bl in
      Textexp_match (e0,brs), 
      res
  | Pextexp_map (e0,bl) ->
      let e0 = type_ext env e0 in
      let (acc,bl) = transl_branches env bl in
      let acc = Cduce_types.Sequence.star acc in
      let t0 = ext_ub env e0.exp_loc acc e0.exp_type in
      let t1 = newvar () in
      let brs = List.map
	(fun (t,p,e) ->
	   let binds,newenv = ext_branch env e.pexp_loc t p t1 in
	   let e = type_ext newenv e in
	   (p,binds,e)) bl in
      let tres = 
	ext_map env loc t0 t1
	  (List.map2 (fun (t,_,_) (_,_,e) -> (t,e.exp_type)) bl brs) in
      Textexp_map (e0,brs), 
      tres
  | Pextexp_xmap (e0,bl) ->
      let e0 = type_ext env e0 in
      let (acc,bl) = transl_branches env bl in
      let t0 = ext_ub env e0.exp_loc Cduce_types.Sequence.any_xtransformable 
	e0.exp_type in
      let t1 = newvar () in
      let brs = List.map
	(fun (t,p,e) ->
	   let binds,newenv = ext_branch env e.pexp_loc t p t1 in
	   let e = type_ext newenv e in
	   (p,binds,e)) bl in
      let tres = 
	ext_xmap env loc acc t0 t1
	  (List.map2 (fun (t,_,_) (_,_,e) -> (t,e.exp_type)) bl brs) in
      Textexp_xmap (e0,brs,t1), 
      tres
  | Pextexp_op (op,el) ->
      let el = List.map (type_ext env) el in
      Textexp_op (op,el), 
      ext_op env loc op (List.map (fun e -> e.exp_type) el)
  | Pextexp_record fl ->
      let fl = transl_record env loc (type_ext env) fl in
      Textexp_record fl,
      ext_record env loc (LabelMap.map (fun e -> e.exp_type) fl)
  | Pextexp_removefield (e,l) ->
      let e = type_ext env e in
      let l = transl_label env loc l in
      Textexp_removefield (e,l),
      ext_removefield env loc e.exp_type l
  | Pextexp_namespace (pr,ns,e) ->
      let e = type_expect (Env.add_namespace pr ns env) e (newvar ()) in
      Textexp_namespace e,
      e.exp_type
  | Pextexp_from_ml e ->
      let t = newvar () in
      let e = type_expect env e t in
      let t = ext_from_ml env loc t in
      Textexp_from_ml e, t
  | Pextexp_to_ml e ->
      let t = new_global_var () in  (* Prevent generalization *)
      let e = type_ext env e in
      ext_to_ml env loc t e.exp_type;
      Textexp_to_ml e,
      t
  | Pextexp_check (e,t) ->
      let e = type_ext env e in
      let t = transl_ext_type env t in
      Textexp_check e,
      t
      

let annot env sexp e = 
  if !extmode then (register_ext_annot env sexp e; e)
  else (
    if sexp.pexp_ext 
    then { e with exp_type = ext_supertype env e.exp_loc e.exp_type }
    else e
  )

let subtype_loc = ref Location.none

let () =
  Ctype.ext_subtype := 
    (fun env t -> 
       if !extmode then anyext_var else ext_subtype env !subtype_loc t);
  Ctype.ext_supertype := 
    (fun env t -> 
       if !extmode then anyext_var else ext_supertype env !subtype_loc t)
